from fastapi import Request, HTTPException
from starlette.responses import JSONResponse
from popbl_chassis.security.securityprovider import SecurityProvider


class AutenticationMiddleware:
    valid_routes = None
    ask_for_keys: bool= None
    generate_keys: bool = None
    role: str = None
    security_service: SecurityProvider = None

    @classmethod
    async def create(cls, valid_routes=None, ask_for_keys: bool = None, generate_keys: bool = None,role: str = None, security_service: SecurityProvider = None):
        self = AutenticationMiddleware()
        self.valid_routes = valid_routes
        self.ask_for_keys = ask_for_keys
        self.generate_keys = generate_keys
        self.security_service = security_service
        self.role = role
        return self
    
    async def dispatch(self, request: Request, call_next):
        for route_method, route_path in self.valid_routes:
            if route_path == request.url.path and (route_method is None or route_method == request.method):
                return await call_next(request)
        try:
            auth = request.headers.get("Authorization")
            if not auth or not auth.startswith("Bearer "):
                return JSONResponse(status_code=401, content={"detail": "Not authenticated"})

            token = auth.split(" ")[1]
            payload = self.security_service.decode_token(token)

            if self.security_service.validar_fecha_expiracion(payload):
                return JSONResponse(status_code=401, content={"detail": "Token has expired"})

            if not self.security_service.validate_role(payload, self.role):
                return JSONResponse(status_code=403, content={"detail": "Insufficient permissions"})

            response = await call_next(request)
            return response

        except HTTPException as exc:
            return JSONResponse(status_code=exc.status_code, content={"detail": exc.detail})
        except Exception as exc:
            print(exc)
            return JSONResponse(status_code=500, content={"detail": "Internal Server Error"})



import pybreaker
import requests
from requests.exceptions import RequestException
from popbl_chassis.discovery.consulprovider import ConsulProvider


class CircuitBreaker:

    consul_provider: ConsulProvider = None
    logger = None
    circuit_breaker: pybreaker.CircuitBreaker = None

    @classmethod
    async def create(cls, logger, consul_provider: ConsulProvider):
        self = CircuitBreaker()
        self.logger = logger
        self.consul_provider = consul_provider
        self.circuit_breaker = pybreaker.CircuitBreaker(fail_max=3, reset_timeout=10)
        return self

    def breaker_call(self, service_name):
        try:
            return self.circuit_breaker.call(self.check_healthcheck, service_name)
        except pybreaker.CircuitBreakerError as cbe:
            self.logger.error(f"Circuit Breaker is open for {service_name}: {cbe}")
            raise
        except Exception as e:
            self.logger.error(f"Error during healthcheck for {service_name}: {e}")
            raise

    def check_healthcheck(self, service_name):
        # Get service address and port from Consul
        ret = self.consul_provider.get_consul_service(f"_{service_name}._tcp")
        if not ret["Address"] or not ret["Port"]:
            self.logger.error(f"Service {service_name} not found in Consul")
            raise ValueError(f"Service {service_name} not found in Consul")

        # Make the request to the /health endpoint
        try:
            response = requests.get(f"http://{ret['Address']}:{ret['Port']}/health", timeout=5)

            # Raise an error if the status code is not 200
            if response.status_code != 200:
                self.logger.warning(f"Healthcheck for {service_name} returned {response.status_code}")
                raise ValueError(f"Healthcheck for {service_name} returned {response.status_code}")

            return True

        except RequestException as re:
            self.logger.error(f"Connection error with {service_name}: {re}")
            raise ValueError(f"Connection error with {service_name}: {re}")



import asyncio
import json
import logging

import aio_pika
from aio_pika import ExchangeType

class LoggerProvider:

    connection: aio_pika.robust_connection = None
    channel: aio_pika.channel = None
    exchange_name = None
    exchange: aio_pika.robust_exchange = None
    rabbitmq_host = None
    rabbitmq_user = None
    rabbitmq_password = None
    logger: logging.Logger = None

    @classmethod
    async def create(cls, rabbitmq_host, rabbitmq_user, rabbitmq_password, logger: logging.Logger):
        self = LoggerProvider()
        self.rabbitmq_host = rabbitmq_host
        self.rabbitmq_user = rabbitmq_user
        self.rabbitmq_password = rabbitmq_password
        self.logger = logger
        await self.subscribe_channel()
        return self

    async def subscribe_channel(self):
        retries = 5
        for attempt in range(retries):
            try:
                self.connection = await aio_pika.connect_robust(
                    host=self.rabbitmq_host,
                    virtualhost='/',
                    login=self.rabbitmq_user,
                    password=self.rabbitmq_password
                )
                self.channel = await self.connection.channel()
                self.exchange_name = 'log'
                self.exchange = await self.channel.declare_exchange(name=self.exchange_name, type=ExchangeType.TOPIC, durable=True)
                self.logger.info("Connection and channel established successfully.")
                break
            except Exception as e:
                self.logger.error(f"Connection failed: {e}")
                if attempt < retries - 1:
                    self.logger.info("Retrying in 5 seconds...")
                    await asyncio.sleep(5)
                else:
                    self.logger.error("All retry attempts failed")
                    raise  # Important to raise an exception here to prevent further execution with a null connection

    async def publish(self, message_body, routing_key):
        if not self.connection or not self.channel:
            self.logger.error("No active connection or channel to publish message.")
            return
        self.logger.info(f"Publishing message to exchange {self.exchange_name} with routing key {routing_key}")
        await self.exchange.publish(
            aio_pika.Message(
                body=message_body.encode(),
                content_type="text/plain"
            ),
            routing_key=routing_key)



import asyncio
import logging.config

import aio_pika
from aio_pika import IncomingMessage, ExchangeType


class EventProvider:

    connection: aio_pika.robust_connection = None
    channel : aio_pika.channel= None
    exchange_name = None
    exchange : aio_pika.robust_exchange = None
    rabbitmq_host = None
    rabbitmq_user = None
    rabbitmq_password = None
    logger : logging.Logger = None

    @classmethod
    async def create(cls,rabbitmq_host, rabbitmq_user, rabbitmq_password, logger: logging.Logger, exchange_name:str='events'):
        self=EventProvider()
        self.rabbitmq_host = rabbitmq_host
        self.rabbitmq_user = rabbitmq_user
        self.rabbitmq_password = rabbitmq_password
        self.logger = logger
        self.exchange_name = exchange_name
        await self.subscribe_channel(exchange_name)
        return self


    async def subscribe_channel(self, exchange_name):
        retries = 5
        for attempt in range(retries):
            try:
                self.connection = await aio_pika.connect_robust(
                    host=self.rabbitmq_host,
                    virtualhost='/',
                    login=self.rabbitmq_user,
                    password=self.rabbitmq_password
                )
            except Exception as e:
                self.logger.error(f"Connection failed: {e}")
                if attempt < retries - 1:
                    self.logger.info(f"Retrying in {5} seconds...")
                    await asyncio.sleep(5)
                else:
                    self.logger.error("All retry attempts failed")
        self.logger.info("Connection established")
        # Create a channel
        self.logger.info("Connected to RabbitMQ")
        self.channel = await self.connection.channel()
        self.logger.info("Channel created")
        self.logger.info("Exchange created")
        self.exchange = await self.channel.declare_exchange(name=exchange_name, type=ExchangeType.TOPIC, durable=True)

    async def subscribe_to_service(self, queue_name:str, routing_key:str,callback: callable([IncomingMessage])):
        # Create a queue
        queue = await self.channel.declare_queue(name=queue_name, exclusive=True)
        # Bind the queue to the exchange
        await queue.bind(exchange=self.exchange_name, routing_key=routing_key)
        # Set up a message consumer
        async with queue.iterator() as queue_iter:
            async for message in queue_iter:
                self.logger.info("Received message en event service order.payed" )
                await callback(message)

    async def publish(self,message_body, routing_key):
        # Publish the message to the exchange
        self.logger.info("Publishing message to exchange %s with routing key %s", self.exchange_name, routing_key)
        await self.exchange.publish(
            aio_pika.Message(
                body=message_body.encode(),
                content_type="text/plain",
                delivery_mode=aio_pika.DeliveryMode.PERSISTENT                
            ),
            routing_key=routing_key)

    async def publish_to_exchange(self, message, routing_key, exchange_name):
        await self.subscribe_channel(exchange_name)
        self.logger.info("Sending log to exchange %s with routing key %s", exchange_name, routing_key)
        await self.publish(message, routing_key)
    
        


import logging

import consul
import dns.resolver
import ifaddr


class ConsulProvider:
    
    consul_port: int= 8500
    consul_dns_port: int= 8600
    consul_host: str= None
    consul_instance: consul = None
    consul_resolver: dns.resolver = None
    ip: str= None
    logger: logging.Logger = None
    
    @classmethod
    async def create(cls, consul_host, logger):
        self = ConsulProvider()
        self.consul_host = consul_host
        self.logger = logger
        self.create_instance()
        self.create_resolver()
        self.set_ip()        
        return self
    
    def set_ip(self):
        ip = self.get_adapter_ip("eth0")  
        self.logger.info(f"service ip is {ip}")
        if ip is None:
            ip = "127.0.0.1"
        self.ip = ip
    
    def create_instance(self):
        self.consul_instance = consul.Consul(
            host=self.consul_host,
            port=self.consul_port
        )
        self.consul_instance.kv.put("aas_example_variable", "aas_example_value")
        
    def create_resolver(self):
        self.consul_resolver = dns.resolver.Resolver(configure=False)
        self.consul_resolver.port = self.consul_dns_port
        self.consul_resolver.nameservers = [str(self.consul_host)]

    @staticmethod
    def get_adapter_ip(nice_name):
        adapters = ifaddr.get_adapters()

        for adapter in adapters:
            if adapter.nice_name == nice_name and len(adapter.ips) > 0:
                return adapter.ips[0].ip

        return None

    def register_consul_service(self, service_name, service_port, service_id):
        self.logger.debug(f"Registering {service_name} service ({service_id})")
        self.consul_instance.agent.service.register(
            name=service_name,
            service_id=service_id,
            address=self.ip,
            port=service_port,
            tags=["python", "microservice", "aas"],
            check={
                "http": 'http://{host}:{port}/health'.format(
                    host=self.ip,
                    port=service_port,
                    service_name=service_name
                ),
                "interval": '10s'
            }
        )
        self.logger.info(f"Registered {service_name} service ({service_id})")
        
        
    def get_consul_service(self, service_name):
        ret = {
            "Address": None,
            "Port": None
        }
        try:
            #  srv_results = consul_dns_resolver.query("{}.service.consul".format(service_name), "srv")
    
            srv_results = self.consul_resolver.resolve(
                "{}.service.consul".format(service_name),
                "srv"
            )  # SRV DNS query
            srv_list = srv_results.response.answer  # PORT - target_name relation
            a_list = srv_results.response.additional  # IP - target_name relation
    
            # DNS returns a list of replicas, supposedly sorted using Round Robin. We always get the 1st element: [0]
            srv_replica = srv_list[0][0]
            port = srv_replica.port
            target_name = srv_replica.target
    
            # From all the IPs, get the one with the chosen target_name
            for a in a_list:
                if a.name == target_name:
                    ret['Address'] = a[0]
                    ret['Port'] = port
                    break
    
        except dns.exception.DNSException as e:
            self.logger.error("Could not get service url: {}".format(e))
        return ret

    def get_consul_key_value_item(self,key):
        """Get consul item value for the given key. It only works for string items!"""
        index, data = self.consul_instance.kv.get(key)
        value = None
        if data and data['Value']:
            value = data['Value'].decode('utf-8')
        return key, value


    def get_consul_service_catalog(self):
        """List al consul services"""
        return self.consul_instance.catalog.services()
    
    
    def get_consul_service_replicas(self):
        """Get all services including replicas"""
        return self.consul_instance.agent.services()
    
    
    def unregister_service_from_consul(self, service_id):
        self.logger.debug(f"Unregistering service ({service_id})")
        self.consul_instance.agent.service.deregister(service_id)
        self.logger.info(f"Unregistered service ({service_id})")


from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession


class DbUtils:
    
    session: AsyncSession= None
    
    @classmethod
    async def create(cls, session: AsyncSession):
        self= DbUtils()
        self.session= session
        return self

    async def get_list(self, model):
        """Retrieve a list of elements from database"""
        result = await self.session.execute(select(model))
        item_list = result.unique().scalars().all()
        return item_list


    async def get_list_statement_result(self, stmt):
        """Execute given statement and return list of items."""
        result = await self.session.execute(stmt)
        item_list = result.unique().scalars().all()
        return item_list


    async def get_element_statement_result(self, stmt):
        """Execute statement and return a single items"""
        result = await self.session.execute(stmt)
        item = result.scalar_one_or_none()
        return item


    async def get_element_by_id(self, model, element_id):
        """Retrieve any DB element by id."""
        if element_id is None:
            return None
        element = await self.session.get(model, element_id)
        return element

    # DELETE
    async def delete_element_by_id(self, model, element_id):
        """Delete any DB element by id."""
        element = await self.get_element_by_id(model, element_id)
        if element is not None:
            await self.session.delete(element)
            await self.session.commit()
        return element

    
    
    

        



import json
import os
from datetime import datetime, timedelta

import jwt
import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from popbl_chassis.discovery.consulprovider import ConsulProvider


class SecurityProvider:
    public_key=None
    private_key=None
    private_pem=None
    public_pem=None
    client_microservice_ip=None
    client_microservice_port=None
    consul_provider: ConsulProvider = None

    @classmethod
    async def create(cls, generate_keys: bool, ask_for_keys: bool, consul_service: ConsulProvider):
        self = SecurityProvider()
        if generate_keys:
            self.generate_keys()
            self.load_public_key()
            self.load_private_key()
            
        if ask_for_keys: 
            self.consul_provider = consul_service
            
        return self
    
    def is_public_key_loaded(self):
        return self.public_key is not None

    def get_public_key(self):
        service= self.consul_provider.get_consul_service("client")
        print(f"service_response:{service}")
        if service["Address"] is None or service["Port"] is None:
            raise ValueError("Client service not found.")
        response= requests.get(f"http://{service["Address"]}:{service["Port"]}/client/get/key")
        data = json.loads(response.text.strip('"').replace("\n", "\n"))
        self.public_key = data['public_key']
        return self.public_key

    @staticmethod
    def load_private_key():
        try:
            with open("private_key.pem", "rb") as key_file:
                return key_file.read()
        except FileNotFoundError:
            raise FileNotFoundError("Private key file not found.")
        except Exception as e:
            raise RuntimeError(f"An error occurred while loading the private key: {e}")

    @staticmethod
    def load_public_key():
        try:
            with open("public_key.pem", "rb") as key_file:
                return key_file.read()
        except FileNotFoundError:
            raise FileNotFoundError("Public key file not found.")
        except Exception as e:
            raise RuntimeError(f"An error occurred while loading the public key: {e}")


    def generate_keys(self):
        if os.path.exists("private_key.pem") and os.path.exists("public_key.pem"):
            print("Keys already exist. Skipping key generation.")
            return

        self.private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())
        self.private_pem = self.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )
        with open("private_key.pem", "wb") as f:
            f.write(self.private_pem)

        self.public_key = self.private_key.public_key()
        self.public_pem = self.public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        with open("public_key.pem", "wb") as f:
            f.write(self.public_pem)

        print("Keys generated successfully.")


    def create_token(self,data: dict, expires_delta: timedelta = timedelta(hours=3)):
        """Creates a JWT token with the given data and expiration."""
        private_key = self.load_private_key()
        data.update({"exp": datetime.utcnow() + expires_delta})
        headers = {
            "alg": "RS256",
            "typ": "JWT"
        }

        return jwt.encode(data, private_key, algorithm="RS256", headers=headers)


    def decode_token(self,token: str):
        """Decodes a JWT token and returns its payload."""
        
        if self.public_key is None:
            self.public_key = self.load_public_key()
            
        try:
            payload = jwt.decode(token, self.public_key, algorithms=["RS256"])
            return payload
        except jwt.ExpiredSignatureError:
            raise ValueError("Token has expired.")
        except jwt.InvalidTokenError:
            raise ValueError("Invalid token.")

    @staticmethod
    def validate_role(payload: dict, role: str) -> bool:
        """Checks if the user has admin privileges."""
        return payload.get("role") == role or payload.get("role") == "admin"

    @staticmethod
    def validar_fecha_expiracion(payload: dict) -> bool:
        # Obtiene la fecha de expiración del token
        exp_timestamp_str = payload.get("fecha_expiracion")
        if not exp_timestamp_str:
            return False

        exp_timestamp_datetime = datetime.fromisoformat(exp_timestamp_str)
        exp_timestamp = exp_timestamp_datetime.timestamp()
        exp_datetime = datetime.utcfromtimestamp(exp_timestamp)

        # Comprueba si el token ha expirado
        return exp_datetime <= datetime.utcnow()


from setuptools import setup, find_packages

setup(
    name="team3_pobl_chassis",
    version="1.0.1",
    packages=find_packages(),
    install_requires=[],
    author="macc-team3",
    author_email="gorka.etxebarria@alumni.mondragon.edu",
    description="A reusable library for microservices",
)
